package com.nisum.ecomorder.controller;

import com.fasterxml.jackson.core.JsonProcessingException;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.github.tomakehurst.wiremock.WireMockServer;
import com.nisum.ecomorder.client.CustomerClient;
import com.nisum.ecomorder.client.ProductClient;
import com.nisum.ecomorder.model.Address;
import com.nisum.ecomorder.model.Customer;
import com.nisum.ecomorder.model.Order;
import com.nisum.ecomorder.model.Product;
import com.nisum.ecomorder.model.dto.OrderLineDto;
import com.nisum.ecomorder.model.dto.PlaceOrderRequestDto;
import com.nisum.ecomorder.repository.OrderRepository;
import org.junit.jupiter.api.*;
import org.junit.jupiter.api.TestInstance.Lifecycle;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.test.context.SpringBootTest;
import org.springframework.boot.test.web.client.TestRestTemplate;
import org.springframework.boot.web.server.LocalServerPort;
import org.springframework.cloud.contract.wiremock.AutoConfigureHttpClient;
import org.springframework.http.HttpStatus;
import org.springframework.http.ResponseEntity;
import org.springframework.test.context.ActiveProfiles;

import java.util.Arrays;

import static com.github.tomakehurst.wiremock.client.WireMock.*;
import static org.junit.jupiter.api.Assertions.*;

@SpringBootTest(webEnvironment = SpringBootTest.WebEnvironment.RANDOM_PORT)
@TestInstance(Lifecycle.PER_CLASS)
@ActiveProfiles("test")
@AutoConfigureHttpClient
public class OrderControllerIntegrationTest {

    @LocalServerPort
    private int port;
    private String baseUrl;
    private final String randomId = "someHexString";
    private final WireMockServer productMockServer = new WireMockServer(8088);
    private final WireMockServer customerMockServer = new WireMockServer(8099);


    @Autowired
    OrderRepository orderRepository;

    @Autowired
    private TestRestTemplate restTemplate;

    @Autowired
    ObjectMapper objectMapper;

    @Autowired
    ProductClient productClient;

    @Autowired
    CustomerClient customerClient;

    @BeforeAll
    public void setUp() {
        initializeDb();
        baseUrl = "http://localhost:" + port;
        productMockServer.start();
        customerMockServer.start();
    }

    @AfterEach
    void after() {
        productMockServer.resetAll();
        customerMockServer.resetAll();
    }

    @AfterAll
    void clean() {
        productMockServer.shutdown();
        customerMockServer.shutdown();
    }

    @Test
    void testProductServer() throws JsonProcessingException {
        createStubForGetProductById();
        Product productById = productClient.getProductById(randomId);
        assertEquals(randomId, productById.getId());
        assertEquals("Bean bag", productById.getName());
    }

    @Test
    void testCustomerServer() throws JsonProcessingException {
        createStubForGetCustomerById();
        Customer customerById = customerClient.getCustomerById(123L);
        assertEquals(123L, customerById.getId());
        assertEquals("Ashok", customerById.getFirstName());
    }

    @Test
    void placeOrder_success() throws JsonProcessingException {
        PlaceOrderRequestDto placeOrderRequestObject = getPlaceOrderRequestObject();
        createStubForGetCustomerById();
        createStubForGetProductById();

        ResponseEntity<Order> orderResponseEntity = this.restTemplate.postForEntity(baseUrl + "/place-order", placeOrderRequestObject, Order.class);
        assertEquals(HttpStatus.CREATED, orderResponseEntity.getStatusCode());
        Order orderResponse = orderResponseEntity.getBody();
        System.out.println(orderResponse);
        assertNotNull(orderResponse);
        assertNotNull(orderResponse.getId());
        assertEquals("Ashok Kumar", orderResponse.getCustomerName());
        orderResponse.getLineItems().forEach(orderLine -> assertNotNull(orderLine.getId()));
        assertEquals(4800D, orderResponse.getOrderTotal());

    }


    private void initializeDb() {
        orderRepository.deleteAll();
    }


    private void createStubForGetProductById() throws JsonProcessingException {
        Product product = new Product(randomId, "Bean bag", "test description", 1200D);
        String responseJson = objectMapper.writeValueAsString(product);
        productMockServer.stubFor(
                get(urlEqualTo("/id/" + randomId))
                        .willReturn(
                                aResponse()
                                        .withHeader("Content-Type", "application/json")
                                        .withStatus(200)
                                        .withBody(responseJson)
                        )
        );
    }

    private void createStubForGetCustomerById() throws JsonProcessingException {
        Customer customer = new Customer(123L, "Ashok", "Kumar", "ashokk@nisum.com", "9999999988");
        String responseJson = objectMapper.writeValueAsString(customer);
        customerMockServer.stubFor(
                get(urlEqualTo("/id/123"))
                        .willReturn(
                                aResponse()
                                        .withHeader("Content-Type", "application/json")
                                        .withStatus(200)
                                        .withBody(responseJson)
                        )
        );
    }

    private PlaceOrderRequestDto getPlaceOrderRequestObject() {
        PlaceOrderRequestDto placeOrderRequestDto = new PlaceOrderRequestDto();
        placeOrderRequestDto.setCustomerId(123L);
        placeOrderRequestDto.setEmail("ashokk@nisum.com");

        Address shippingAddress = new Address();
        shippingAddress.setAddressLine1("Jawaharnagar colony");
        shippingAddress.setAddressLine2("Ameerpet");
        shippingAddress.setCity("Hyderabad");
        shippingAddress.setState("TS");
        shippingAddress.setCountry("IND");
        shippingAddress.setZipCode("500001");

        placeOrderRequestDto.setShipTo(shippingAddress);

        OrderLineDto orderLineDto1 = new OrderLineDto();
        orderLineDto1.setProductId(randomId);
        orderLineDto1.setQuantity(2);

        OrderLineDto orderLineDto2 = new OrderLineDto();
        orderLineDto2.setProductId(randomId);
        orderLineDto2.setQuantity(2);

        placeOrderRequestDto.setLineItems(Arrays.asList(orderLineDto1, orderLineDto2));

        return placeOrderRequestDto;
    }

}
